@@ -43,6 +43,9 @@
 #include "jsdbgapi.h"
 #include "jsscript.h"
 
+// Default GC zeal level for new JS contexts.
+static uint8 PYM_defaultGCZeal;
+
 // This is the default throw hook for pydermonkey-owned JS contexts,
 // when they've defined one in Python.
 static JSTrapStatus
@@ -109,10 +112,11 @@ PYM_reportError(JSContext *cx, const char *message, JSErrorReport *report)
                          NULL, NULL);
     else
       PyErr_Warn(NULL, message);
-  } else
-    // TODO: Not sure if this will ever get called, but we should know
-    // if it is.
-    PyErr_Warn(NULL, "A JS error was reported.");
+  } else {
+    PyErr_SetString(PYM_error, message);
+    if (JS_IsExceptionPending(cx))
+      JS_ClearPendingException(cx);
+  }
 }
 
 static int
@@ -127,6 +131,14 @@ PYM_traverse(PYM_JSContextObject *self, visitproc visit, void *arg)
 static int
 PYM_clear(PYM_JSContextObject *self)
 {
+  // Dropping our owned reference to the runtime means that it
+  // could disappear, so we're going to have to destroy our own
+  // context before we do that.
+  if (self->cx) {
+    JS_DestroyContext(self->cx);
+    self->cx = NULL;
+  }
+
   Py_CLEAR(self->opCallback);
   Py_CLEAR(self->throwHook);
   Py_CLEAR(self->runtime);
@@ -139,16 +151,34 @@ PYM_JSContextDealloc(PYM_JSContextObject *self)
   if (self->weakrefs)
     PyObject_ClearWeakRefs((PyObject *) self);
   PyObject_GC_UnTrack(self);
-
-  if (self->cx) {
-    JS_DestroyContext(self->cx);
-    self->cx = NULL;
-  }
-
   PYM_clear(self);
   PyObject_GC_Del(self);
 }
 
+static int
+PYM_pyObjectToPropertyJsval(PYM_JSContextObject *context,
+                            PyObject *object,
+                            jsval *rval)
+{
+  if ((PyInt_Check(object) && !INT_FITS_IN_JSVAL(PyInt_AS_LONG(object))) ||
+      PyLong_Check(object)) {
+    PyErr_SetString(PyExc_ValueError,
+                    "Integer property value out of range.");
+    return -1;
+  }
+
+  if (PYM_pyObjectToJsval(context, object, rval) == -1)
+    return -1;
+
+  if (!(JSVAL_IS_STRING(*rval) || JSVAL_IS_INT(*rval))) {
+    PyErr_SetString(PyExc_TypeError,
+                    "Property must be a string or integer.");
+    return -1;
+  }
+
+  return 0;
+}
+
 static PyObject *
 PYM_getRuntime(PYM_JSContextObject *self, PyObject *args)
 {
@@ -301,6 +331,55 @@ PYM_getPendingException(PYM_JSContextObject *self, PyObject *args)
   return PYM_jsvalToPyObject(self, exception);
 }
 
+PyObject *
+PYM_setDefaultGCZeal(PyObject *self, PyObject *args)
+{
+  int level;
+  if (!PyArg_ParseTuple(args, "i", &level))
+    return NULL;
+
+  if (level >= 0 && level <= 2) {
+    PYM_defaultGCZeal = level;
+    Py_RETURN_NONE;
+  }
+
+  PyErr_SetString(PyExc_ValueError, "Level must be between 0 and 2.");
+  return NULL;
+}
+
+static PyObject *
+PYM_getVersion(PYM_JSContextObject *self, PyObject *args)
+{
+  PYM_SANITY_CHECK(self->runtime);
+
+  const char *version = JS_VersionToString(JS_GetVersion(self->cx));
+  return PyString_FromString(version);
+}
+
+static PyObject *
+PYM_setGCZeal(PYM_JSContextObject *self, PyObject *args)
+{
+#ifdef JS_GC_ZEAL
+  PYM_SANITY_CHECK(self->runtime);
+
+  int level;
+  if (!PyArg_ParseTuple(args, "i", &level))
+    return NULL;
+
+  if (level >= 0 && level <= 2) {
+    JS_SetGCZeal(self->cx, level);
+    Py_RETURN_NONE;
+  }
+
+  PyErr_SetString(PyExc_ValueError, "Level must be between 0 and 2.");
+  return NULL;
+#else
+  PyErr_SetString(PyExc_NotImplementedError,
+                  "Spidermonkey built without support for JS_SetGCZeal()");
+  return NULL;
+#endif
+}
+
 static PyObject *
 PYM_getObjectPrivate(PYM_JSContextObject *self, PyObject *args)
 {
@@ -363,16 +442,39 @@ PYM_clearObjectPrivate(PYM_JSContextObject *self, PyObject *args)
   Py_RETURN_NONE;
 }
 
+static PyObject *
+PYM_newArrayObject(PYM_JSContextObject *self, PyObject *args)
+{
+  PYM_SANITY_CHECK(self->runtime);
+
+  JSObject *array = JS_NewArrayObject(self->cx, 0, NULL);
+  if (array == NULL) {
+    PYM_jsExceptionToPython(self);
+    return NULL;
+  }
+
+  return PYM_jsvalToPyObject(self, OBJECT_TO_JSVAL(array));
+}
+
 static PyObject *
 PYM_newObject(PYM_JSContextObject *self, PyObject *args)
 {
   PYM_SANITY_CHECK(self->runtime);
   PyObject *privateObj = NULL;
+  PYM_JSObject *proto = NULL;
 
-  if (!PyArg_ParseTuple(args, "|O", &privateObj))
+  if (!PyArg_ParseTuple(args, "|OO!", &privateObj,
+                        &PYM_JSObjectType, &proto))
     return NULL;
 
-  JSObject *obj = PYM_JS_newObject(self->cx, privateObj);
+  JSObject *jsProto = NULL;
+
+  if (proto) {
+    PYM_ENSURE_RUNTIME_MATCH(self->runtime, proto->runtime);
+    jsProto = proto->obj;
+  }
+
+  JSObject *obj = PYM_JS_newObject(self->cx, privateObj, jsProto, NULL);
   if (obj == NULL) {
     PyErr_SetString(PYM_error, "PYM_JS_newObject() failed");
     return NULL;
@@ -388,21 +490,31 @@ PYM_hasProperty(PYM_JSContextObject *self, PyObject *args)
 {
   PYM_SANITY_CHECK(self->runtime);
   PYM_JSObject *object;
-  char *buffer = NULL;
-  int size;
+  PyObject *property;
 
-  if (!PyArg_ParseTuple(args, "O!es#", &PYM_JSObjectType, &object,
-                        "utf-16", &buffer, &size))
+  if (!PyArg_ParseTuple(args, "O!O", &PYM_JSObjectType, &object,
+                        &property))
     return NULL;
 
-  PYM_UTF16String str(buffer, size);
-
   PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
 
+  jsval propertyVal;
+  if (PYM_pyObjectToPropertyJsval(self, property, &propertyVal) == -1)
+    return NULL;
+
   JSBool hasProperty;
   JSBool result;
-  result = JS_HasUCProperty(self->cx, object->obj, str.jsbuffer,
-                            str.jslen, &hasProperty);
+  if (JSVAL_IS_INT(propertyVal)) {
+    result = JS_HasElement(self->cx, object->obj,
+                           JSVAL_TO_INT(propertyVal),
+                           &hasProperty);
+  } else {
+    JSString *str = JSVAL_TO_STRING(propertyVal);
+    result = JS_HasUCProperty(self->cx, object->obj,
+                              JS_GetStringChars(str),
+                              JS_GetStringLength(str),
+                              &hasProperty);
+  }
 
   if (!result) {
     PYM_jsExceptionToPython(self);
@@ -419,22 +531,73 @@ PYM_getProperty(PYM_JSContextObject *self, PyObject *args)
 {
   PYM_SANITY_CHECK(self->runtime);
   PYM_JSObject *object;
-  char *buffer = NULL;
-  int size;
+  PyObject *property;
 
-  if (!PyArg_ParseTuple(args, "O!es#", &PYM_JSObjectType, &object,
-                        "utf-16", &buffer, &size))
+  if (!PyArg_ParseTuple(args, "O!O", &PYM_JSObjectType, &object,
+                        &property))
     return NULL;
 
-  PYM_UTF16String str(buffer, size);
-
   PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
 
+  jsval propertyVal;
+  if (PYM_pyObjectToPropertyJsval(self, property, &propertyVal) == -1)
+    return NULL;
+
   jsval val;
   JSBool result;
+
   Py_BEGIN_ALLOW_THREADS;
-  result = JS_GetUCProperty(self->cx, object->obj, str.jsbuffer,
-                            str.jslen, &val);
+  if (JSVAL_IS_INT(propertyVal)) {
+    result = JS_GetElement(self->cx, object->obj,
+                           JSVAL_TO_INT(propertyVal), &val);
+  } else {
+    JSString *str = JSVAL_TO_STRING(propertyVal);
+    result = JS_GetUCProperty(self->cx, object->obj,
+                              JS_GetStringChars(str),
+                              JS_GetStringLength(str),
+                              &val);
+  }
+  Py_END_ALLOW_THREADS;
+
+  if (!result) {
+    PYM_jsExceptionToPython(self);
+    return NULL;
+  }
+
+  return PYM_jsvalToPyObject(self, val);
+}
+
+static PyObject *
+PYM_lookupProperty(PYM_JSContextObject *self, PyObject *args)
+{
+  PYM_SANITY_CHECK(self->runtime);
+  PYM_JSObject *object;
+  PyObject *property;
+
+  if (!PyArg_ParseTuple(args, "O!O", &PYM_JSObjectType, &object,
+                        &property))
+    return NULL;
+
+  PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
+
+  jsval propertyVal;
+  if (PYM_pyObjectToPropertyJsval(self, property, &propertyVal) == -1)
+    return NULL;
+
+  jsval val;
+  JSBool result;
+
+  Py_BEGIN_ALLOW_THREADS;
+  if (JSVAL_IS_INT(propertyVal)) {
+    result = JS_LookupElement(self->cx, object->obj,
+                              JSVAL_TO_INT(propertyVal), &val);
+  } else {
+    JSString *str = JSVAL_TO_STRING(propertyVal);
+    result = JS_LookupUCProperty(self->cx, object->obj,
+                                 JS_GetStringChars(str),
+                                 JS_GetStringLength(str),
+                                 &val);
+  }
   Py_END_ALLOW_THREADS;
 
   if (!result) {
@@ -464,6 +627,14 @@ PYM_initStandardClasses(PYM_JSContextObject *self, PyObject *args)
 
   PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
 
+  if (JS_GetGlobalObject(self->cx)) {
+    // TODO: This is really just a workaround for issue #3:
+    // http://code.google.com/p/pydermonkey/issues/detail?id=3
+    PyErr_SetString(PYM_error,
+                    "Can't init standard classes on the same context twice.");
+    return NULL;
+  }
+
   if (!JS_InitStandardClasses(self->cx, object->obj)) {
     PyErr_SetString(PYM_error, "JS_InitStandardClasses() failed");
     return NULL;
@@ -588,33 +759,194 @@ PYM_evaluateScript(PYM_JSContextObject *self, PyObject *args)
   return pyRval;
 }
 
+static PyObject *
+PYM_isArrayObject(PYM_JSContextObject *self, PyObject *args)
+{
+  PYM_SANITY_CHECK(self->runtime);
+  PYM_JSObject *object;
+
+  if (!PyArg_ParseTuple(args, "O!", &PYM_JSObjectType, &object))
+    return NULL;
+
+  PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
+
+  if (JS_IsArrayObject(self->cx, object->obj))
+    Py_RETURN_TRUE;
+  Py_RETURN_FALSE;
+}
+
+static PyObject *
+PYM_enumerate(PYM_JSContextObject *self, PyObject *args)
+{
+  PYM_SANITY_CHECK(self->runtime);
+  PYM_JSObject *object;
+
+  if (!PyArg_ParseTuple(args, "O!", &PYM_JSObjectType, &object))
+    return NULL;
+
+  PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
+
+  JSIdArray *idArray = JS_Enumerate(self->cx, object->obj);
+  if (idArray == NULL) {
+    PYM_jsExceptionToPython(self);
+    return NULL;
+  }
+
+  PyObject *tuple = PyTuple_New(idArray->length);
+  if (tuple == NULL) {
+    JS_DestroyIdArray(self->cx, idArray);
+    return NULL;
+  }
+
+  for (int i = 0; i < idArray->length; i++) {
+    jsval val;
+    if (!JS_IdToValue(self->cx, idArray->vector[i], &val)) {
+      Py_DECREF(tuple);
+      JS_DestroyIdArray(self->cx, idArray);
+      PYM_jsExceptionToPython(self);
+      return NULL;
+    }
+
+    PyObject *item = PYM_jsvalToPyObject(self, val);
+    if (item == NULL) {
+      Py_DECREF(tuple);
+      JS_DestroyIdArray(self->cx, idArray);
+      return NULL;
+    }
+
+    // Note that this function "steals" a reference to item, so
+    // we don't need to decrement its reference count.
+    PyTuple_SET_ITEM(tuple, i, item);
+  }
+
+  JS_DestroyIdArray(self->cx, idArray);
+  return tuple;
+}
+
+static PyObject *
+PYM_setProperty(PYM_JSContextObject *self, PyObject *args)
+{
+  PYM_SANITY_CHECK(self->runtime);
+  PYM_JSObject *object;
+  PyObject *property;
+  PyObject *value;
+
+  if (!PyArg_ParseTuple(args, "O!OO", &PYM_JSObjectType, &object,
+                        &property, &value))
+    return NULL;
+
+  PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
+
+  // TODO: Do we really need this, or do we just want to root a single var?
+  JSAutoLocalRootScope localRootScope(self->cx);
+  jsval propertyVal;
+  if (PYM_pyObjectToPropertyJsval(self, property, &propertyVal) == -1)
+    return NULL;
+
+  jsval jsValue;
+  if (PYM_pyObjectToJsval(self, value, &jsValue) == -1)
+    return NULL;
+
+  JSBool result;
+  if (JSVAL_IS_INT(propertyVal)) {
+    result = JS_SetElement(self->cx, object->obj,
+                           JSVAL_TO_INT(propertyVal),
+                           &jsValue);
+  } else {
+    JSString *str = JSVAL_TO_STRING(propertyVal);
+    result = JS_SetUCProperty(self->cx, object->obj,
+                              JS_GetStringChars(str),
+                              JS_GetStringLength(str),
+                              &jsValue);
+  }
+
+  if (!result) {
+    PYM_jsExceptionToPython(self);
+    return NULL;
+  }
+
+  return PYM_jsvalToPyObject(self, jsValue);
+}
+
+static PyObject *
+PYM_deleteProperty(PYM_JSContextObject *self, PyObject *args)
+{
+  PYM_SANITY_CHECK(self->runtime);
+  PYM_JSObject *object;
+  PyObject *property;
+
+  if (!PyArg_ParseTuple(args, "O!O", &PYM_JSObjectType, &object,
+                        &property))
+    return NULL;
+
+  PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
+
+  jsval propertyVal;
+  if (PYM_pyObjectToPropertyJsval(self, property, &propertyVal) == -1)
+    return NULL;
+
+  jsval jsValue;
+  JSBool result;
+  if (JSVAL_IS_INT(propertyVal)) {
+    result = JS_DeleteElement2(self->cx, object->obj,
+                               JSVAL_TO_INT(propertyVal),
+                               &jsValue);
+  } else {
+    JSString *str = JSVAL_TO_STRING(propertyVal);
+    result = JS_DeleteUCProperty2(self->cx, object->obj,
+                                  JS_GetStringChars(str),
+                                  JS_GetStringLength(str),
+                                  &jsValue);
+  }
+
+  if (!result) {
+    PYM_jsExceptionToPython(self);
+    return NULL;
+  }
+
+  return PYM_jsvalToPyObject(self, jsValue);
+}
+
 static PyObject *
 PYM_defineProperty(PYM_JSContextObject *self, PyObject *args)
 {
   PYM_SANITY_CHECK(self->runtime);
   PYM_JSObject *object;
+  PyObject *property;
   PyObject *value;
-  char *name = NULL;
-  int namelen;
 
-  if (!PyArg_ParseTuple(args, "O!es#O", &PYM_JSObjectType, &object,
-                        "utf-16", &name, &namelen, &value))
+  if (!PyArg_ParseTuple(args, "O!OO", &PYM_JSObjectType, &object,
+                        &property, &value))
     return NULL;
 
-  PYM_UTF16String str(name, namelen);
-  jsval jsValue;
-
   PYM_ENSURE_RUNTIME_MATCH(self->runtime, object->runtime);
 
+  // TODO: Do we really need this, or do we just want to root a single var?
+  JSAutoLocalRootScope localRootScope(self->cx);
+  jsval propertyVal;
+  if (PYM_pyObjectToPropertyJsval(self, property, &propertyVal) == -1)
+    return NULL;
+
+  jsval jsValue;
   if (PYM_pyObjectToJsval(self, value, &jsValue) == -1)
     return NULL;
 
-  if (!JS_DefineUCProperty(self->cx, object->obj, str.jsbuffer,
-                           str.jslen, jsValue, NULL, NULL,
-                           JSPROP_ENUMERATE)) {
-    // TODO: There's probably an exception pending on self->cx,
-    // what should we do about it?
-    PyErr_SetString(PYM_error, "JS_DefineProperty() failed");
+  JSBool result;
+  if (JSVAL_IS_INT(propertyVal)) {
+    result = JS_DefineElement(self->cx, object->obj,
+                              JSVAL_TO_INT(propertyVal),
+                              jsValue, NULL, NULL, JSPROP_ENUMERATE);
+  } else {
+    JSString *str = JSVAL_TO_STRING(propertyVal);
+    result = JS_DefineUCProperty(self->cx, object->obj,
+                                 JS_GetStringChars(str),
+                                 JS_GetStringLength(str),
+                                 jsValue, NULL, NULL,
+                                 JSPROP_ENUMERATE);
+  }
+
+  if (!result) {
+    PYM_jsExceptionToPython(self);
     return NULL;
   }
 
@@ -747,6 +1079,8 @@ static PyMethodDef PYM_JSContextMethods[] = {
    "Get the JavaScript runtime associated with this context."},
   {"get_stack", (PyCFunction) PYM_getStack, METH_VARARGS,
    "Get the current stack for the context."},
+  {"new_array_object", (PyCFunction) PYM_newArrayObject, METH_VARARGS,
+   "Create a new JavaScript Array object."},
   {"new_object", (PyCFunction) PYM_newObject, METH_VARARGS,
    "Create a new JavaScript object."},
   {"init_standard_classes",
@@ -771,13 +1105,25 @@ static PyMethodDef PYM_JSContextMethods[] = {
   {"new_function",
    (PyCFunction) PYM_newFunction, METH_VARARGS,
    "Creates a new function callable from JS."},
+  {"is_array_object", (PyCFunction) PYM_isArrayObject, METH_VARARGS,
+   "Returns whether or not the given JavaScript object is an array."},
+  {"enumerate", (PyCFunction) PYM_enumerate, METH_VARARGS,
+   "Returns a tuple of all a JavaScript object's enumerable properties."},
   {"define_property",
    (PyCFunction) PYM_defineProperty, METH_VARARGS,
-   "Defines a property on an object."},
+   "Defines a property on a JavaScript object."},
+  {"set_property",
+   (PyCFunction) PYM_setProperty, METH_VARARGS,
+   "Sets a property on a JavaScript object."},
   {"get_property", (PyCFunction) PYM_getProperty, METH_VARARGS,
    "Gets the given property for the given JavaScript object."},
+  {"lookup_property", (PyCFunction) PYM_lookupProperty, METH_VARARGS,
+   "Gets the stored value of the given property for the given "
+   "JavaScript object."},
   {"has_property", (PyCFunction) PYM_hasProperty, METH_VARARGS,
    "Returns whether the given JavaScript object has the given property."},
+  {"delete_property", (PyCFunction) PYM_deleteProperty, METH_VARARGS,
+   "Deletes the given property on the given object."},
   {"gc", (PyCFunction) PYM_gc, METH_VARARGS,
    "Performs garbage collection on the context's runtime."},
   {"set_operation_callback", (PyCFunction) PYM_setOperationCallback,
@@ -788,6 +1134,10 @@ static PyMethodDef PYM_JSContextMethods[] = {
   {"trigger_operation_callback", (PyCFunction) PYM_triggerOperationCallback,
    METH_VARARGS,
    "Triggers the operation callback for the context."},
+  {"get_version", (PyCFunction) PYM_getVersion, METH_VARARGS,
+   "Returns the JS version of the context."},
+  {"set_gc_zeal", (PyCFunction) PYM_setGCZeal, METH_VARARGS,
+   "Sets the frequency of garbage collection."},
   {"get_object_private", (PyCFunction) PYM_getObjectPrivate, METH_VARARGS,
    "Returns the private Python object stored in the JavaScript object."},
   {"clear_object_private", (PyCFunction) PYM_clearObjectPrivate, METH_VARARGS,
@@ -868,9 +1218,7 @@ PYM_newJSContextObject(PYM_JSRuntimeObject *runtime, JSContext *cx)
   JS_SetErrorReporter(cx, PYM_reportError);
 
 #ifdef JS_GC_ZEAL
-  // TODO: Consider exposing JS_SetGCZeal() to Python instead of
-  // hard-coding it here.
-  JS_SetGCZeal(cx, 2);
+  JS_SetGCZeal(cx, PYM_defaultGCZeal);
 #endif
 
   PyObject_GC_Track((PyObject *) context);
