@@ -7,6 +7,8 @@ import threading
 
 import pydermonkey
 
+pydermonkey.set_default_gc_zeal(2)
+
 class PydermonkeyTests(unittest.TestCase):
     def setUp(self):
         self._teardowns = []
@@ -24,11 +26,13 @@ class PydermonkeyTests(unittest.TestCase):
     def _clearOnTeardown(self, obj):
         self._teardowns.append(obj)
 
-    def _evaljs(self, code):
-        rt = pydermonkey.Runtime()
-        cx = rt.new_context()
-        obj = cx.new_object()
-        cx.init_standard_classes(obj)
+    def _evaljs(self, code, cx=None, obj=None):
+        if cx is None:
+            rt = pydermonkey.Runtime()
+            cx = rt.new_context()
+        if obj is None:
+            obj = cx.new_object()
+            cx.init_standard_classes(obj)
         return cx.evaluate_script(obj, code, '<string>', 1)
 
     def _execjs(self, code):
@@ -58,7 +62,7 @@ class PydermonkeyTests(unittest.TestCase):
         self.assertTrue(was_raised)
 
     def testVersionIsCorrect(self):
-        # Really hackish way of importing values from setup.py.
+        # Really hackish way of importing values from metadata.py.
         import os
 
         mydir = os.path.dirname(__file__)
@@ -70,6 +74,63 @@ class PydermonkeyTests(unittest.TestCase):
 
         self.assertEqual(pydermonkey.__version__, metadata.VERSION)
 
+    def testDeletePropertyWorks(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.define_property(obj, 'foo', 1)
+        self.assertEqual(cx.delete_property(obj, 'foo'), True)
+        self.assertEqual(cx.delete_property(obj, 'foo'), True)
+        self.assertEqual(cx.get_property(obj, 'foo'),
+                         pydermonkey.undefined)
+
+    def testSetPropertyWorks(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        self.assertEqual(cx.set_property(obj, 'blah', 5), 5)
+        self.assertEqual(cx.set_property(obj, 3, 2), 2)
+        self.assertEqual(cx.set_property(obj, u'blah\u2026', 5), 5)
+        self.assertEqual(cx.get_property(obj, 3), 2)
+
+    def testSetPropertyReturnsDifferentValue(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.init_standard_classes(obj)
+        o2 = self._evaljs("({set blah() { return 5; }})", cx, obj)
+        self.assertEqual(cx.set_property(o2, 'blah', 3), 5)
+
+    def testSetPropertyRaisesExceptionOnReadOnly(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.init_standard_classes(obj)
+        o2 = self._evaljs("({get blah() { return 5; }})", cx, obj)
+        self.assertRaises(
+            pydermonkey.error,
+            cx.set_property,
+            o2, 'blah', 3
+            )
+        self.assertEqual(self.last_exception.args[0],
+                         "setting a property that has only a getter")
+
+    def testDefinePropertyRaisesNoExceptionOnReadOnly(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.init_standard_classes(obj)
+        o2 = self._evaljs("({get blah() { return 5; }})", cx, obj)
+        cx.define_property(o2, 'blah', 3)
+        self.assertEqual(cx.get_property(o2, 'blah'), 3)
+
+    def testLookupPropertyWorks(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.init_standard_classes(obj)
+        o2 = self._evaljs("({foo: 1, get blah() { return 5; }})", cx, obj)
+        self.assertEqual(cx.lookup_property(o2, 'bar'),
+                         pydermonkey.undefined)
+        self.assertEqual(cx.lookup_property(o2, 'foo'), 1)
+        self.assertEqual(cx.lookup_property(o2, 'blah'), True)
+        self.assertEqual(cx.get_property(o2, 'blah'), 5)
+        self.assertEqual(cx.lookup_property(o2, 'blah'), True)
+
     def testSyntaxErrorsAreRaised(self):
         for run in [self._evaljs, self._execjs]:
             self.assertRaises(pydermonkey.error, run, '5f')
@@ -109,6 +170,72 @@ class PydermonkeyTests(unittest.TestCase):
         self.assertTrue(pc >= 0 and pc < len(buffer(script)))
         self.assertEqual(stack['caller']['caller']['caller'], None)
 
+    def testNewObjectTakesProto(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.define_property(obj, 5, "foo")
+        obj2 = cx.new_object(None, obj)
+        self.assertEqual(cx.get_property(obj2, 5), "foo")
+
+    def testInitStandardClassesWorksTwice(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.init_standard_classes(obj)
+        obj2 = cx.new_object()
+
+        # TODO: This is really just a workaround for issue #3:
+        # http://code.google.com/p/pydermonkey/issues/detail?id=3
+        self.assertRaises(
+            pydermonkey.error,
+            cx.init_standard_classes,
+            obj2
+            )
+        self.assertEqual(
+            self.last_exception.args[0],
+            "Can't init standard classes on the same context twice."
+            )             
+
+    def testNewArrayObjectWorks(self):
+        cx = pydermonkey.Runtime().new_context()
+        array = cx.new_array_object()
+        self.assertTrue(cx.is_array_object(array))
+
+    def testIsArrayWorks(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        self.assertFalse(cx.is_array_object(obj))
+        array = cx.evaluate_script(obj, '[1]', '<string>', 1)
+        self.assertTrue(cx.is_array_object(array))
+
+    def testEnumerateWorks(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.evaluate_script(obj, "var blah = 1; var foo = 2; this[0] = 5;",
+                           "<string>", 1)
+        self.assertEqual(cx.enumerate(obj), ("blah", "foo", 0))
+
+    def testBigArrayIndicesRaiseValueError(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        self.assertRaises(
+            ValueError,
+            cx.define_property,
+            obj, 2 ** 30, 'foo'   # Should be a PyInt object.
+            )
+        self.assertEqual(self.last_exception.args[0],
+                         "Integer property value out of range.")
+
+    def testReallyBigArrayIndicesRaiseValueError(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        self.assertRaises(
+            ValueError,
+            cx.define_property,
+            obj, sys.maxint + 1, 'foo'   # Should be a PyLong object.
+            )
+        self.assertEqual(self.last_exception.args[0],
+                         "Integer property value out of range.")
+
     def testScriptHasFilenameMember(self):
         cx = pydermonkey.Runtime().new_context()
         script = cx.compile_script('foo', '<string>', 1)
@@ -191,6 +318,42 @@ class PydermonkeyTests(unittest.TestCase):
         gc.collect()
         self.assertEqual(wcx(), None)
 
+    def testKeyboardInterruptStopsScript(self):
+        # Let's be super-evil and have multiple interleavings of the JS
+        # stack with the Python stack.
+
+        def opcb(cx):
+            raise KeyboardInterrupt()
+
+        cx = pydermonkey.Runtime().new_context()
+        cx.set_operation_callback(opcb)
+        obj = cx.new_object()
+        cx.init_standard_classes(obj)
+
+        def func(cx, this, args):
+            cx.evaluate_script(
+                this,
+                'try { while (1) {} } catch (e) {}',
+                '<string>', 1
+                )
+
+        cx.define_property(obj,
+                           'func',
+                           cx.new_function(func, 'func'))
+
+        def watchdog():
+            time.sleep(0.1)
+            cx.trigger_operation_callback()
+
+        thread = threading.Thread(target = watchdog)
+        thread.start()
+
+        self.assertRaises(
+            KeyboardInterrupt,
+            cx.evaluate_script,
+            obj, 'while (1) { func(); }', '<string>', 1
+            )
+
     def testOperationCallbackIsCalled(self):
         def opcb(cx):
             raise Exception("stop eet!")
@@ -223,6 +386,16 @@ class PydermonkeyTests(unittest.TestCase):
                                     '(function(){})', '<string>', 1)
         self.assertFalse(jsfunc.is_python)
 
+    def testRecreatedJsWrappedPythonFuncHasIsPythonTrue(self):
+        def foo(cx, this, args):
+            pass
+
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.define_property(obj, 'foo',
+                           cx.new_function(foo, foo.__name__))
+        self.assertTrue(cx.get_property(obj, 'foo').is_python)
+
     def testJsWrappedPythonFuncHasIsPythonTrue(self):
         def foo(cx, this, args):
             pass
@@ -463,6 +636,12 @@ class PydermonkeyTests(unittest.TestCase):
                                                    'hai2u("blah\x00 ")'),
                          u"blah\x00 o hai")
 
+    def testJsWrappedPythonFunctionReturnsUndefined(self):
+        def hai2u(cx, this, args):
+            return pydermonkey.undefined
+        self.assertEqual(self._evalJsWrappedPyFunc(hai2u, 'hai2u()'),
+                         pydermonkey.undefined)
+
     def testJsWrappedPythonFunctionReturnsString(self):
         def hai2u(cx, this, args):
             return "o hai"
@@ -483,6 +662,14 @@ class PydermonkeyTests(unittest.TestCase):
                           hai2u, 'hai2u()')
         self.assertEqual(self.last_exception.args[0], u"blarg")
 
+    def testJsWrappedPythonFunctionThrowsJsException2(self):
+        def hai2u(cx, this, args):
+            cx.evaluate_script(this, 'throw "blarg"', '<string>', 1)
+        self.assertRaises(pydermonkey.error,
+                          self._evalJsWrappedPyFunc,
+                          hai2u, 'hai2u()')
+        self.assertEqual(self.last_exception.args[0], u"blarg")
+
     def testJsWrappedPythonFunctionThrowsPyException(self):
         thecx = []
         def hai2u(cx, this, args):
@@ -536,6 +723,36 @@ class PydermonkeyTests(unittest.TestCase):
         self.assertEqual(self._evalJsWrappedPyFunc(hai2u, 'hai2u()'),
                          2147483647)
 
+    def testDefinePropertyWorksWithIntegers(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.define_property(obj, 0, 'test')
+        self.assertEqual(cx.evaluate_script(obj, "this[0]", '<string>', 1),
+                         'test')
+
+    def testGetPropertyDoesNotWorkWithFloats(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        self.assertRaises(TypeError,
+                          cx.get_property,
+                          obj, 0.534)
+        self.assertEqual(self.last_exception.args[0],
+                         'Property must be a string or integer.')
+
+    def testHasPropertyWorksWithIntegers(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        self.assertEqual(cx.has_property(obj, 0), False)
+        cx.define_property(obj, 0, 'hi')
+        self.assertEqual(cx.has_property(obj, 0), True)
+
+    def testGetPropertyWorksWithIntegers(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        cx.init_standard_classes(obj)
+        array = cx.evaluate_script(obj, "['test']", '<string>', 1)
+        self.assertEqual(cx.get_property(array, 0), 'test')
+
     def testHasPropertyWorks(self):
         cx = pydermonkey.Runtime().new_context()
         obj = cx.new_object()
@@ -760,6 +977,19 @@ class PydermonkeyTests(unittest.TestCase):
         self.assertEqual(cx.evaluate_script(obj, '5+3', '<string>', 1),
                          8)
 
+    def testTypeConversionNotImplementedMentionsTypeName(self):
+        cx = pydermonkey.Runtime().new_context()
+        obj = cx.new_object()
+        self.assertRaises(
+            NotImplementedError,
+            cx.define_property,
+            obj, 'hi', 2 ** 91
+            )
+        self.assertEqual(
+            self.last_exception.args[0],
+            "Data type conversion not implemented for type 'long'."
+            )
+
     def testCallFunctionRaisesErrorOnBadFuncArgs(self):
         cx = pydermonkey.Runtime().new_context()
         obj = cx.new_object()
@@ -817,6 +1047,20 @@ class PydermonkeyTests(unittest.TestCase):
             )
         self.assertEqual(cx.call_function(thisArg, obj, (1,2)), 6)
 
+    def testGetVersionWorks(self):
+        # Note that this will change someday.
+        self.assertEqual(pydermonkey.Runtime().new_context().get_version(),
+                         "1.8")
+
+    def testSetGCZealWorks(self):
+        cx = pydermonkey.Runtime().new_context()
+        for i in range(3):
+            pydermonkey.set_default_gc_zeal(i)
+            cx.set_gc_zeal(i)
+        for i in [-1, 3]:
+            self.assertRaises(ValueError, cx.set_gc_zeal, i)
+            self.assertRaises(ValueError, pydermonkey.set_default_gc_zeal, i)
+
     def testEvaluateReturnsTrue(self):
         self.assertTrue(self._evaljs('true') is True)
 
